#!/usr/bin/env python
"""Point of execution for play.

Configures module path and libraries and then calls lib.main.main.

"""

import getopt
import os
import sys
import textwrap
import traceback


def prepare_path():
    """Load the source directory in to sys.path.

    This is only necessary when running a source distribution. Binary
    distributions ensure that the module path contains the modules anyway.

    """
    if os.path.exists("src"):
        sys.path.insert(1, "src")
        #sys.path.insert(1, "squirtle")
        sys.path.insert(1, "cocos_squirtle")
        sys.path.insert(1, "cocos")
        sys.path.insert(1, "yaml")
        sys.path.insert(1, "lepton")
        #sys.path.insert(1, "pyknic")


def fix_avbin_loader():
    """Monkey patch pyglet.lib to load AVBin from the current directory on
    OS X if it is found.

    WARNING: this is currently required to make sure that the py2app
    distribution launches correctly on systems without AVBin.

    """
    from pyglet.lib import loader
    find_library = loader.find_library
    def fixed_find_library(name):
        if loader.platform == "darwin" and os.path.exists("libavbin.dylib") \
                and name == "avbin":
            return "libavbin.dylib"
        return find_library(name)
    loader.find_library = fixed_find_library


def set_debug(debug):
    """Set the debug flag in the constants module to the given value.

    :Parameters:
        `debug` : bool
            The value to set.

    """
    try:
        import constants
        constants.DEBUG = debug
    except ImportError, exc:
        print "error: could not find source directory"
        sys.exit(1)


def parse_args():
    """Parse command line arguments.

    For details of what the options do see USAGE_MESSAGE (defined below).

    """
    import config as config
    import constants

    print constants.DEBUG
    ## Construct the usage message.
    USAGE_MESSAGE = textwrap.dedent("""\
    usage: %s [OPTIONS]...
    Mandatory arguments for long options are required for short options too.

      -h, --help                    print this usage message
      -v, --version                 print version information
    """ % os.path.basename(os.path.join(".", sys.argv[0])))

    ## Add debug options to usage message.
    if constants.DEBUG: USAGE_MESSAGE += textwrap.dedent("""\

    Debug options
    -------------
      -p, --profile                 run in the profiler
    """)

    ## Utility functions
    def usage(exit_code):
        print USAGE_MESSAGE
        sys.exit(exit_code)
    def version():
        print u"v%s" % constants.VERSION
        sys.exit()
    def assert_debug():
        if not constants.DEBUG:
            usage(2)

    ## Parse arguments.
    try:
        opts, args = getopt.getopt(sys.argv[1:], "hvp", ["help", "version", "profile"])
    except getopt.GetoptError, exc:
        usage(2)

    ## No additional arguments.
    if len(args) > 0:
        usage(2)

    ## Process options.
    for opt, arg in opts:
        if opt in ("-h", "--help"):
            usage(0)
        elif opt in ("-v", "--version"):
            version()
        elif opt in ("-p", "--profile"):
            assert_debug()
            config.profile = True
        else:
            usage(2)

def setup_logging():
    import logging
    import constants
    
    
    if constants.DEBUG:
        loglevel = logging.DEBUG
    else:
        loglevel = logging.CRITICAL
    #loglevel = logging.INFO    
    logging.basicConfig(level = loglevel,
                        format   ='%(asctime)s %(name)-12s %(levelname)-8s %(message)s',
                        datefmt  ='%H:%M:%S', filename = constants.LOGFILE, filemode = 'w')
                        
    console = logging.StreamHandler()
    console.setLevel(loglevel)

    formatter = logging.Formatter('%(name)-12s %(levelname)-8s %(message)s')
    console.setFormatter(formatter)
    logging.getLogger('').addHandler(console)
    
     
def run(debug=False):
    """Run the game.

    Prepares the top level before launching the game. The aim is to make moving
    from development to production simple, and to ensure differences between
    distributions are ironed out before starting the game.

    :Parameters:
        `debug` : bool
            The value to set for constants.DEBUG.

    """
    prepare_path()
    fix_avbin_loader()
    set_debug(debug)
    parse_args()
    setup_logging()
    
    try:
        from PIL import TgaImagePlugin
        import main
        main.main()
        #import test_level_bg
        #test_level_bg.run()
        #import test_world
        #test_world.run()
        #import test_tiles
        #test_tiles.run()
    except Exception, exc:
        import constants
        open(constants.TMP_LOC+constants.CONFIG_NAME+"_error.log", "w").write(traceback.format_exc())
        raise


if __name__ == "__main__":

    # Change to the game directory
    os.chdir(os.path.dirname(os.path.join(".", sys.argv[0])))

    # Start the actual game
    run()
